"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.PartitionManager = void 0;
const events_1 = require("events");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const partition_1 = require("./partition");
/**
 * The PartitionManager is responsible for maintaining a list of partitions for the given Kafka topic.
 * It will route incoming messages to the appropriate partition for the messages.
 * @internal
 */
class PartitionManager extends events_1.EventEmitter {
    constructor(factory, consumer, logger, config, listenForConsumerErrors = true) {
        super();
        this.factory = factory;
        this.consumer = consumer;
        this.logger = logger;
        this.config = config;
        this.partitions = new Map();
        // Start rebalancing until we receive the first rebalanced message
        this.isRebalancing = true;
        this.stopped = false;
        // Place new Kafka messages into our processing queue
        this.consumer.on("data", (message) => {
            this.process(message);
        });
        this.consumer.on("rebalancing", (partitions) => {
            this.rebalancing(partitions);
        });
        this.consumer.on("rebalanced", (partitions) => {
            this.rebalanced(partitions);
        });
        if (listenForConsumerErrors) {
            this.consumer.on("error", (error, errorData) => {
                if (this.stopped) {
                    server_services_telemetry_1.Lumberjack.info("Consumer.onError: PartitionManager already stopped, not emitting error again", { error, ...errorData });
                    return;
                }
                this.emit("error", error, errorData);
            });
            this.consumer.on("checkpoint_success", (partitionId, queuedMessage, retries, latency) => {
                if (this.sampleMessages(100)) {
                    server_services_telemetry_1.Lumberjack.info(`Kafka checkpoint successful`, {
                        msgOffset: queuedMessage.offset,
                        topic: queuedMessage.topic,
                        msgPartition: queuedMessage.partition,
                        retries,
                        latency,
                    });
                }
            });
            this.consumer.on("checkpoint_error", (partitionId, queuedMessage, retries, latency, ex) => {
                server_services_telemetry_1.Lumberjack.error(`Kafka checkpoint failed`, {
                    msgOffset: queuedMessage.offset,
                    topic: queuedMessage.topic,
                    msgPartition: queuedMessage.partition,
                    retries,
                    latency,
                }, ex);
            });
        }
    }
    async stop() {
        this.stopped = true;
        this.logger?.info("Stop requested");
        server_services_telemetry_1.Lumberjack.info("Stop requested");
        // Drain all pending messages from the partitions
        const partitionsStoppedP = [];
        for (const [, partition] of this.partitions) {
            const stopP = partition.drain();
            partitionsStoppedP.push(stopP);
        }
        await Promise.all(partitionsStoppedP);
        // Then stop them all
        for (const [, partition] of this.partitions) {
            partition.close(server_services_core_1.LambdaCloseType.Stop);
        }
        this.partitions.clear();
        this.removeAllListeners();
    }
    process(message) {
        if (this.stopped) {
            return;
        }
        if (this.isRebalancing) {
            this.logger?.info(`Ignoring ${message.topic}:${message.partition}@${message.offset} due to pending rebalance`);
            server_services_telemetry_1.Lumberjack.info(`Ignoring ${message.topic}:${message.partition}@${message.offset} due to pending rebalance`);
            return;
        }
        const partition = this.partitions.get(message.partition);
        if (!partition) {
            this.emit("error", `Received message for untracked partition ${message.topic}:${message.partition}@${message.offset}`);
            return;
        }
        partition.process(message);
    }
    /**
     * Called when rebalancing starts
     * Note: The consumer may decide to only emit "rebalanced" if it wants to skip closing existing partitions
     * @param partitions - Assigned partitions before the rebalance
     */
    rebalancing(partitions) {
        this.logger?.info(`Rebalancing partitions: ${JSON.stringify(partitions)}`);
        server_services_telemetry_1.Lumberjack.info(`Rebalancing partitions: ${JSON.stringify(partitions)}`);
        this.isRebalancing = true;
        for (const [id, partition] of this.partitions) {
            this.logger?.info(`Closing partition ${id} due to rebalancing`);
            server_services_telemetry_1.Lumberjack.info(`Closing partition ${id} due to rebalancing`);
            partition.close(server_services_core_1.LambdaCloseType.Rebalance);
        }
        this.partitions.clear();
    }
    /**
     * Called when rebalanced occurs
     * @param partitions - Assigned partitions after the rebalance.
     * May contain partitions that have been previously assigned to this consumer
     */
    rebalanced(partitions) {
        if (this.stopped) {
            return;
        }
        this.isRebalancing = false;
        const partitionsMap = new Map(partitions.map((partition) => [partition.partition, partition]));
        // close and remove existing partitions that are no longer assigned
        const existingPartitions = Array.from(this.partitions);
        for (const [id, partition] of existingPartitions) {
            if (!partitionsMap.has(id)) {
                this.logger?.info(`Closing partition ${id} due to rebalancing`);
                server_services_telemetry_1.Lumberjack.info(`Closing partition ${id} due to rebalancing`);
                partition.close(server_services_core_1.LambdaCloseType.Rebalance);
                this.partitions.delete(id);
            }
        }
        // create new partitions
        for (const partition of partitions) {
            if (this.partitions.has(partition.partition)) {
                // this partition already exists
                // it must have existed before the rebalance
                continue;
            }
            this.logger?.info(`Creating ${partition.topic}: Partition ${partition.partition}, Offset ${partition.offset} due to rebalance`);
            server_services_telemetry_1.Lumberjack.info(`Creating ${partition.topic}: Partition ${partition.partition}, Offset ${partition.offset} due to rebalance`);
            const newPartition = new partition_1.Partition(partition.partition, this.factory, this.consumer, this.logger, this.config);
            // Listen for error events to know when the partition has stopped processing due to an error
            newPartition.on("error", (error, errorData) => {
                if (this.stopped) {
                    server_services_telemetry_1.Lumberjack.info("Partition.onError: PartitionManager already stopped, not emitting error again", { error, ...errorData });
                    return;
                }
                server_services_telemetry_1.Lumberjack.verbose("Emitting error from partitionManager, partition error event");
                this.emit("error", error, errorData);
            });
            this.partitions.set(partition.partition, newPartition);
        }
    }
    sampleMessages(numberOfMessagesPerTrace) {
        return this.getRandomInt(numberOfMessagesPerTrace) === 0;
    }
    getRandomInt(range) {
        return Math.floor(Math.random() * range);
    }
}
exports.PartitionManager = PartitionManager;
//# sourceMappingURL=partitionManager.js.map